﻿
#pragma once

#include <cstring>
#include <cmath>
#include <exception>
#include <iostream>
#include <vector>
#include <algorithm>
#include <functional>
#include "clx/thread.h"
#include "clx/mutex.h"
#include "othelloBoard.hpp"

//オセロプレーヤーの抽象クラス
class IOthelloPlayer {
public:
	//思考ルーチン。srcに一手分進めたものを代入する
	virtual void Place(Board &src) = 0;
};

//人間のプレーヤー。入力を監視しながら打つ
class UserOthelloPlayer: public IOthelloPlayer {
private:
	int pos;
	clx::mutex m;
	//入力を監視して取り出す
	int PeekPos() {
		{
			clx::mutex::scoped_lock lock(m);
			pos = -1;
		}
		int res = -1;
		while (res == -1) {
			clx::mutex::scoped_lock lock(m);
			if (pos >= 0) {
				res = pos;
				pos = -1;
			}
		}
		return res;
	}
public:
	//入力を監視して、入力があればその場所に打ってsrcに代入
	void Place(Board &src) {
		while (1) {
			int pos = PeekPos();	
			if (src.GetPlacable(pos)) {
				src.Place(pos);
				return;
			}
		}
	}
	//外部から入力を与える
	void Message(int x, int y) {
		clx::mutex::scoped_lock lock(m);
		pos = x + 8 * y;
	}
};

//平均と分散を求めるためのクラス
class StatCalculator {
private:
	int count;
	double sum;
	double sum2;
	//安全に除算をするためにcountを1以上にして返す
	int SafeCount() {
		return std::max(1, count); 
	}
public:
	StatCalculator() {
		Clear();
	}
	//平均
	double Mean() {
		return sum / SafeCount();
	}
	//分散
	double Variance() {
		double m = Mean();
		return sum2 / SafeCount() - m * m;
	}
	//データを与える
	void Add(double a) {
		count += 1;
		sum += a;
		sum2 += a * a;
	}
	//合計
	double Sum() {
		return sum;
	}
	//クリア
	void Clear() {
		count = 0;
		sum = sum2 = 0;
	}
};

//コンピュータのプレイヤー
class OthelloPlayer1: public IOthelloPlayer {
private:	
	static const int MaxExpanded = 600000; //目標展開ノード数
	int maxDepth; //探索の深さ
	Board selectedBoard; //選ばれた手
	int pass; //パスの回数
	int expandedNodeCount; //実際に展開されたノード数
	StatCalculator expandedStat; //枝狩り前の展開されるはずだったノード数の統計
	StatCalculator prunnedStat; //枝狩り適用後の実際に展開されたノード数の統計
	StatCalculator totalExpandedStat; //expandedStatと同じだが、毎回の手番で蓄積する。実験にデータを取るために使った
	StatCalculator totalPrunnedStat; //prunnedStatと同じだが、毎回の手番で蓄積する。実験にデータを取るために使った
	int maxExpandDepth; 
	double parameter1; //評価関数の重み付け
	bool useFixed; //確定石(Fixed)を使うかどうか

	//評価関数。黒から見たものを計算する
	double EvaluateCore(Board a) {
		double e1, e2;
		e1 = e2 = 0;

		for (int i = 0; i < 64; i++) {
			if (a.GetPlacable(i))
				e1 += 1;
			if (a.GetDisk(i) == 0 && (!useFixed || a.GetFixed(i)))
				e2 += 1;
		}
		e1 *= parameter1 * (1.0 /  64);
		e2 *= (1 - parameter1) * (1.0 / 64);
		return e1 + e2;
	}
	//評価関数。黒から見たもの - 白から見たもの
	double Evaluate(Board a) {
		double eval1 = EvaluateCore(a);    
		a.Flip();
		double eval2 = EvaluateCore(a);
		return eval1 - eval2; 
	}
	//予想評価関数値
	struct PredictedEval {
		double eval;
		Board board;		
		friend bool operator <(const PredictedEval &a, const PredictedEval &b) {
			return a.eval > b.eval;
		}
		friend bool operator >(const PredictedEval &a, const PredictedEval &b) {
			return a.eval < b.eval;
		}
	};	
	//思考ルーチン
	double Find(Board src, int depth, double alpha, double beta) {
		maxExpandDepth = std::max(maxExpandDepth, depth);
		expandedNodeCount++;

		BoardIterator it(src);
		if (it.IsEnd()) {
			//打てる手が無い場合
			expandedStat.Add(0);
			prunnedStat.Add(0);
			totalExpandedStat.Add(0);
			totalPrunnedStat.Add(0);

			pass++;
			if (pass == 2) {
				//パスが二回。つまりゲーム終了の場合
				pass = 0;
				int black = src.BlackCount();
				int white = src.WhiteCount();
				double res = black - white;
				//勝ち負けに対応した評価値を返す
				//勝ちは、早くて枚数差が多いほうが良い
				//負けは、遅くて枚数差が少ないほうが良い
				if (res == 0)
					return std::max(alpha, 0.0);
				else if (res < 0)
					return std::max(alpha, res - 64 * (10000 - depth));
				else
					return std::max(alpha, res + 64 * (10000 - depth));
			}
			else {
				//パスの場合
				src.Flip();
				alpha = std::max(alpha, -Find(src, depth + 1, -beta, -alpha));
				if (depth == 0)
					selectedBoard = *it;
				return alpha;
			}
		}	
		//ノードを展開して、予測評価値を計算する
		pass = 0;
		std::vector<PredictedEval> list;
		for (; it.IsEnd() == 0; ++it) {
			PredictedEval eval;
			eval.eval = Evaluate(*it);
			eval.board = *it;
			list.push_back(eval);
		}
		double logSize = std::log((double)list.size());
		expandedStat.Add(logSize);
		totalExpandedStat.Add(logSize);
		//予測評価値でソートする
		std::stable_sort(list.begin(), list.end());

		for (int i = 0; i < list.size(); i++) {
			double e;
			if (depth >= maxDepth) {
				//最後のノードの場合は、予測評価値を評価値とする
				e = list[i].eval;
			}
			else {
				//ノードの展開
				Board temp = list[i].board;
				temp.Flip();
				e = -Find(temp, depth + 1, -beta, -alpha);
			}
			if (e > alpha) {				
				if (depth == 0) 
					selectedBoard = list[i].board;
				alpha = e;
			}	

			//枝狩り
			if (alpha >= beta) {
				prunnedStat.Add(std::log((double)(i + 1)));
				totalPrunnedStat.Add(std::log((double)(i + 1)));
				return beta;
			}
		}
		prunnedStat.Add(logSize);
		totalPrunnedStat.Add(logSize);
		return alpha;
	}
public:
	//コンストラクタ
	OthelloPlayer1(double _parameter1, bool _useFixed) {
		Initialize();
		useFixed = _useFixed;
		parameter1 = _parameter1;
	}
	//Log(枝狩り率)
	double PrunnedRateLog() {
		return totalPrunnedStat.Mean() - totalExpandedStat.Mean();
	}
	//Log(枝狩り率)の分散
	double PrunnedRateVariLog() {
		return totalPrunnedStat.Variance() - totalExpandedStat.Variance();
	}
	double Parameter1() {
		return parameter1;
	}
	bool UseFixed() {
		return useFixed;
	}
	void Initialize() {
		maxDepth = 6;
	}
	//思考ルーチン
	void Place(Board &board) {
		expandedStat.Clear();
		prunnedStat.Clear();
		maxExpandDepth = 0;

		pass = 0;
		expandedNodeCount = 0;
		//探索
		double eval = Find(board, 0, -(1 << 30), (1 << 30));
		board = selectedBoard;			
		
		using namespace std;		

		cout << "  maxDepth(actual/expected): " << maxExpandDepth << "/" << maxDepth << endl;
		cout << "  expandMean: " << std::exp(prunnedStat.Mean()) << "/" << std::exp(expandedStat.Mean()) << endl;
		cout << "  expandedNodes: " << expandedNodeCount << endl;
		cout << "  eval: " << eval << endl;		

		//次回の探索深さを決定する
		maxDepth = (int)(log((double)MaxExpanded) / std::log(1 + std::exp(expandedStat.Mean())));
	}
};

//オセロゲームループ
class Othello {
private:
	clx::mutex m;
	clx::thread *th;
	Board current, current2;
	IOthelloPlayer *player1;
	IOthelloPlayer *player2;
	int turn;
	struct OthelloThread {
		Othello *parent;
		void operator() () {
			parent->threadFunc();
		}
	};
	//メインループ
	void threadFunc() {
		using namespace std;

		int pass = 0;
		IOthelloPlayer *p1 = player1;
		IOthelloPlayer *p2 = player2;
		while (pass < 2) { //パスが二回になったら終了
			if (current.PlacableCount() > 0) {		
				//プレイヤーに打たせる
				p1->Place(current);
				clx::mutex::scoped_lock lock(m);
				current2 = current;
				if (p1 == player2)
					current2.Flip();
				pass = 0;
			}
			else {
				//パス
				pass++;
				if (GetTurn() == 0)
					cout << "pass: black" << endl;
				else
					cout << "pass: white" << endl;
			}
			//プレイヤーを交代、盤面を反転
			std::swap(p1, p2);
			current.Flip();
			{
				clx::mutex::scoped_lock lock(m);
				turn ^= 1;

				cout << "black, white: (" << current2.BlackCount()
					<< ", " << current2.WhiteCount() << ")" << endl;
			}
		}
		using namespace std;
		cout << "terminated" << endl;
	}
public:
	//コンストラクタ
	Othello(IOthelloPlayer *aplayer1, IOthelloPlayer *aplayer2): player1(aplayer1), player2(aplayer2) {
		th = 0;
		NewGame();
	}	
	//ゲームの終了を待つ
	void WaitFor() {
		if (th) {
			th->join();
			th = NULL;
		}
	}
	//新規ゲーム
	void NewGame() {
		WaitFor();

		current = Board();
		current.SetDisk(3 + 8 * 3, 1);
		current.SetDisk(3 + 8 * 4, 0);
		current.SetDisk(4 + 8 * 3, 0);
		current.SetDisk(4 + 8 * 4, 1);
		current2 = current;
		turn = 0;

		OthelloThread a;
		a.parent = this;
		th = new clx::thread(a);
	}
	//現在のゲーム番の状態を取得
	Board GetCurrent() { 	
		clx::mutex::scoped_lock lock(m);
		Board res;		
		res = current2;
		return res;
	}
	//現在の手番を取得
	int GetTurn() {
		clx::mutex::scoped_lock lock(m);
		return turn;
	}
};