﻿
#pragma once

#include <cstring>
#include <cmath>
#include <exception>

/*
placablesは下位2bit使う。0: 黒が置けるかどうか、1: しろが置けるかどうか
disksは下位2bit使う、0: 黒、1: しろ, 2:無し
オセロ盤の位置は、x + 8 * y = 0から63で表す
*/

#define DiskCount (64)

//オセロ盤を直線にスキャンするためのイテレータ
class BoardScanIterator {
private:
	int x, y;
	int dirX, dirY;	
	int *DirTable() {
		static int dirTable[16] = {
			1, 0,
			-1, 0,
			0, 1,
			0, -1,
			1, 1,
			-1, -1,
			1, -1,
			-1, 1
		};
		return dirTable;
	}
	void Initialize(int ax, int ay, int dirIndex) {
		x = ax;
		y = ay;
		dirX = DirTable()[2 * dirIndex + 0];
		dirY = DirTable()[2 * dirIndex + 1];
	}
public:	
	//コンストラクタ
    BoardScanIterator(int ax, int ay, int dirIndex) {		
		Initialize(ax, ay, dirIndex);
	}
	BoardScanIterator(int pos, int dirIndex) {
		Initialize(pos % 8, pos / 8, dirIndex);
	}
	//現在のオセロ盤の位置を返す
    int operator *() const {
        return x + 8 * y;
    };
	//次のマスに進める
    BoardScanIterator& operator ++(){
        x += dirX;
		y += dirY;
        return *this;
    };
	//戻る
	BoardScanIterator& operator --(){
        x -= dirX;
		y -= dirY;
        return *this;
    };
	//オセロ盤からはみ出した(0以外を返す)かどうか
	int IsEnd() {
		return (x + (y << 8)) & (~(7 + (7 << 8))); //小手先の最適化。3clockで実行可能
		//return (x < 0 || 8 <= x || y < 0 || 8 <= y); 
	}
};

//オセロ盤の状態を記憶するクラス
class Board {
private:
	unsigned char disks[DiskCount]; 
	unsigned char placableCache[DiskCount];
	bool fixedCache[DiskCount];
	bool fixedCacheAvailable;
	bool placableCacheAvailable;
	//itで指定された方向に石がひっくり返せるかどうかを返す関数
	bool PlacableDir(BoardScanIterator it) {		
		++it;	
		int count = 0;
		for (; it.IsEnd() == 0; ++it) {
			switch (disks[*it]) {			
			case 0:
				return count > 0;
			case 1:
				count++;
				break;
			case 2:
				return false;
			}
		}
		return false;
	}
	//itで指定された方向に石をひっくり返す関数
	void PlaceDir(BoardScanIterator it) {		
		if (PlacableDir(it)) {
			++it;
			for (; it.IsEnd() == 0 && disks[*it] != 0; ++it) 
				disks[*it] = 0;						
		}
	}
	//diskで指定された種類の石の数を数える関数
	int GetDiskCount(int disk) {
		int res = 0;
		for (int i = 0; i < DiskCount; i++)
			if (GetDisk(i) == disk)
				res++;
		return res;
	}
public:
	//コンストラクタ
	Board() {	
		for (int i = 0; i < DiskCount; i++) {
			disks[i] = 2;
			fixedCache[i] = false;
		}
		placableCacheAvailable = false;
	}
	Board(const Board &src) {
		for (int i = 0; i < DiskCount; i++) {
			disks[i] = src.disks[i];
			placableCache[i] = src.placableCache[i];
			fixedCache[i] = src.fixedCache[i];
		}
		placableCacheAvailable = src.placableCacheAvailable;
		fixedCacheAvailable = src.fixedCacheAvailable;
	}
	//白と黒を逆転させる関数
	void Flip() {
		for (int i = 0; i < DiskCount; i++) {
			if (disks[i] != 2)
				disks[i] ^= 1;
		}
		placableCacheAvailable = false;
	}
	//posの場所に黒の石を置いてルールに基づきひっくり返す関数
	void Place(int pos) {
		for (int i = 0; i < 8; i++)
			PlaceDir(BoardScanIterator(pos, i));
		disks[pos] = 0;
		placableCacheAvailable = false;
		fixedCacheAvailable = false;
	}	
	//比較演算子
	bool operator== (const Board &a) {
		for (int i = 0; i < DiskCount; i++)
			if (disks[i] != a.disks[i])
				return false;
		return true;
	}
	//posの場所に黒の石を置けるかどうかを返す関数
	bool GetPlacable(int pos) {
		if (!placableCacheAvailable) {
			placableCacheAvailable = true;
			for (int i = 0; i < DiskCount; i++)
				placableCache[i] = 0;
		}
		if ((placableCache[pos] & 1) == 0) {
			placableCache[pos] = 1;
			if (disks[pos] == 2)
				for (int i = 0; i < 8; i++)
					if (PlacableDir(BoardScanIterator(pos, i))) {						
						placableCache[pos] |= 2;
						break;
					}
		}

		return (placableCache[pos] & 2) != 0;
	}
	//posの場所の石が確定石かどうかを返す関数
	bool GetFixed(int pos) {
		if (!fixedCacheAvailable) {
			fixedCacheAvailable = true;
			int count = 1;
			while (count > 0) {
				count = 0;

				for (int i = 0; i < DiskCount; i++) {
					if (disks[i] == 2)
						continue;
					bool fixed = true;
					if (!fixedCache[i] && i != 0 && i != 7 && i != 56 && i != 63) {
						for (int j = 0; j < 8; j++) {
							BoardScanIterator it(i, j);
							--it;
							bool flag = true;
							for (; it.IsEnd() == 0 && !fixedCache[*it]; --it)
								if (disks[*it] == 2) {
									flag = false;
									break;
								}
							if (flag)
								continue;
							it = BoardScanIterator(i, j);
							++it;
							for (; it.IsEnd() == 0; ++it)
								if (!fixedCache[*it] || disks[*it] != disks[i]) {
									fixed = false;
									break;
								}
						}
					}
					if (fixedCache[i] != fixed)
						count++;
					fixedCache[i] = fixed;
				}
			}
		}
		return fixedCache[pos];			
	}
	//posの場所の石の種類(黒、白、空き)を返す関数。
	int GetDisk(int pos) {		
		return disks[pos];
	}
	//posの場所に石を置く関数。ひっくり返す処理は無い
	void SetDisk(int pos, int disk) {
		disks[pos] = disk;		
		placableCacheAvailable = false;
		fixedCacheAvailable = false;
	}
	//黒の数
	int BlackCount() {
		return GetDiskCount(0);
	}
	//白の数
	int WhiteCount() {
		return GetDiskCount(1);
	}
	//空きの数
	int EmptyCount() {
		return GetDiskCount(2);
	}
	//黒の石を置ける場所の数
	int PlacableCount() {
		int count = 0;
		for (int i = 0; i < DiskCount; i++)
			if (GetPlacable(i))
				count++;
		return count;
	}
	//黒の確定石の数
	int FixedCount() {
		int count = 0;
		for (int i = 0; i < DiskCount; i++)
			if (GetFixed(i))
				count++;
		return count;
	}
	//ゲームが終了したかどうかを返す関数
	bool Gameover() {
		if (PlacableCount() > 0)
			return false;
		Board temp = *this;
		temp.Flip();
		return temp.PlacableCount() == 0;		
	}
};

//次の打てる手を打ったBoardを次々に返すイテレータ
class BoardIterator  {
private:
	Board board;
	Board current;
	int pos;
	//次に石をおける場所を探して、あれば置いてcurrentに代入。無ければ終了(pos = -1)
	void findNext() {
		while (pos < DiskCount) {
			if (board.GetPlacable(pos)) {
				current = Board(board);
				current.Place(pos);
				pos++;
				return;
			}
			pos++;
		}		
		pos = -1;
	}
public:
    BoardIterator(Board &b): board(b), pos(0) {	
		findNext();
	}
	//現在の値(オセロ盤)を返す
    Board& operator *() {
        return current;
    };
	//次に進める
    BoardIterator& operator ++(){
		findNext();
        return *this;
    };
	//終了したかどうか
	int IsEnd() {
		return (pos == -1);
	}
};